from __future__ import with_statement
import sys
from time import time, sleep

from thread import get_ident, allocate_lock as Lock, start_new_thread
from threading import currentThread as current_thread, _active, _active_limbo_lock, _limbo, _MainThread

from mext.classhacks import get_name
from mext.context import State
from mext.log import Logger


__all__ = [
    'Thread', 'current_thread', 'is_main_thread',
    'Lock', 'RLock', 'Event', 'Queue'
    #'Empty', 'Full', 'PriorityQueue', 'LifoQueue'
]


assert current_thread().name == 'MainThread'
main_thread_id = get_ident()

def is_main_thread():
    return get_ident() == main_thread_id




class RLock(object):
    def __init__(self):
        self.lock = Lock()
        self.owner = None
        self.depth = 0

    def __repr__(self):
        try:
            owner = _active[self.owner].name
        except KeyError:
            owner = self.owner
        return "<%s owner=%r count=%d>" % (self.__class__.__name__, owner, self.count)

    def acquire(self, blocking=True):
        me = get_ident()
        if self.owner == me:
            self.count += 1
            return 1
        rc = self.lock.acquire(blocking)
        if rc:
            self.owner = me
            self.count = 1
        return rc

    __enter__ = acquire

    def release(self):
        if self.owner != get_ident():
            raise RuntimeError("cannot release un-acquired lock")
        self.count = count = self.count - 1
        if not count:
            self.owner = None
            self.lock.release()

    def __exit__(self, t, v, tb):
        self.release()



def _wait_timeout(lock, timeout):
    endtime = time() + timeout
    delay = 0.0005 # initial delay of 1 ms
    while True:
        if lock.acquire(0):
            return True
        delay = min(delay * 2, endtime - time(), .05)
        if delay <= 0:
            break
        sleep(delay)


class Condition(object):
    def __init__(self, lock=None):
        self.lock = lock or Lock()
        self.waiters = []
        self._temp_release = self.lock.release
        self._restore_lock = lambda state: self.lock.acquire()

    def __enter__(self):
        return self.lock.__enter__()

    def __exit__(self, *args):
        return self.lock.__exit__(*args)

    def __repr__(self):
        return "<%s(%d)>" % (self.__class__.__name__, len(self.waiters))

    def _check_locked(self):
        assert self.lock.locked(), "cannot wait on un-acquired lock"

    def wait(self, timeout=None):
        self._check_locked()
        waiter = Lock()
        waiter.acquire()
        self.waiters.append(waiter)
        state = self._temp_release()
        try:
            if timeout is None:
                waiter.acquire()
            else:
                gotit = _wait_timeout(waiter, timeout)
                if not gotit:
                    self._remove_waiter(waiter)
        finally:
            self._restore_lock(state)

    def _remove_waiter(self, waiter):
        try:
            self.waiters.remove(waiter)
        except ValueError:
            pass

    def notify(self, n=1):
        self._check_locked()
        waiters = self.waiters[:n]
        if not waiters:
            return
        for waiter in waiters:
            waiter.release()
            self._remove_waiter(waiter)

    def notify_all(self):
        self.notify(len(self.waiters))


class RCondition(Condition):
    def __init__(self):
        self.lock = RLock()
        self.waiters = []

    def _check_locked(self):
        assert (self.lock.owner == get_ident()), "cannot wait on un-acquired lock"

    def _temp_release(self):
        l = self.lock
        r = l.count, l.owner
        l.count = 0
        l.owner = None
        l.lock.release()
        return r

    def _restore_lock(self, state):
        l = self.lock
        l.lock.acquire()
        l.count, l.owner = state




class Semaphore(object):
    def __init__(self, value=1, verbose=None):
        if value < 0:
            raise ValueError("semaphore initial valuet must be >= 0")
        self.cond = Condition()
        self.value = value

    def acquire(self, blocking=True):
        with self.cond:
            while self.value == 0:
                if not blocking:
                    return False
                self.cond.wait()
            self.value -= 1
            return True

    __enter__ = acquire

    def release(self):
        with self.cond:
            self.value += 1
            self.cond.notify()

    def __exit__(self, t, v, tb):
        self.release()



class Event(object):
    def __init__(self):
        self.cond = Condition()
        self.flag = False

    def is_set(self):
        return self.flag

    def set(self):
        with self.cond:
            self.flag = True
            self.cond.notify_all()

    def clear(self):
        with self.cond:
            self.flag = False

    def wait(self, timeout=None):
        with self.cond:
            if not self.flag:
                self.cond.wait(timeout)






class Thread(object):
    state = None
    log = Logger('threads')
    ignore_errors = False
    #_exc_clear = sys.exc_clear

    def __init__(self, func, *args, **kw):
        self.func = func
        self.name = get_name(func, 'Thread-%s' % func)
        self.args = args
        self.kw = kw

        self.daemon = current_thread().daemon
        self.id = None
        self.evt_started = Event()
        self.stopped = False
        self.cond = Condition()


    def __repr__(self):
        status = "initial"
        if self.evt_started.is_set():
            status = "started"
        if self.stopped:
            status = "stopped"
        if self.daemon:
            status += " daemon"
        if self.id is not None:
            status += " %s" % self.id
        return "<%s(%s, %s)>" % (self.__class__.__name__, self.name, status)




    @property
    def is_current(self):
        return self.id == get_ident()

    @property
    def is_main(self):
        return self.id == main_thread_id

    get = get_current = staticmethod(current_thread)

    @classmethod
    def find(self, id):
        return _active.get(id)






    def start_daemon(self):
        self.daemon = True
        return self.start()

    def start(self):
        if self.state is None:
            self.state = State.clone()
        #self.log.info("Starting new thread %s", self.name)
        if self.evt_started.is_set():
            raise RuntimeError("thread already started")
        with _active_limbo_lock:
            _limbo[self] = self
        try:
            start_new_thread(self._thread_main, ())
        except:
            with _active_limbo_lock:
                del _limbo[self]
            raise
        self.evt_started.wait()
        return self



    def _thread_main(self):
        try:
            try:
                self.id = get_ident()
                self.evt_started.set()
                with _active_limbo_lock:
                    _active[self.id] = self
                    del _limbo[self]
                self._thread_main_inner()
            finally:
                with _active_limbo_lock:
                    with self.cond:
                        self.stopped = True
                        self.cond.notify_all()
                    _active.pop(get_ident(), None)
        except:
            if not self.daemon or sys is not None:
                raise


    def _thread_main_inner(self):
        with self.state:
            try:
                #self.log.info("Thread %s is running", self.name)
                self.func(*self.args, **self.kw)
            except SystemExit:
                raise
            except:
                self._handle_error()
            finally:
                #self.log.debug("Thread stopped")
                # Avoid a refcycle if the thread is running a function with
                # an argument that has a member that points to the thread.
                del self.func, self.args, self.kw
                #self._exc_clear()

    def _handle_error(self):
        if sys is None:
            raise
        exc_info = sys.exc_info()
        if not self.ignore_errors:
            sys.excepthook(*exc_info)
        #elif isinstance(self.ignore_errors, Logger)::
        elif self.ignore_errors == 'log':
            self.log.error("Thread raised an exception", exc_info=True)
        else:
            self.log.debug(u"Thread raised an exception %s", exc_info[1])
        del exc_info

    def join(self, timeout=None):
        if not self.evt_started.is_set():
            raise RuntimeError("cannot join thread before it is started")
        if self.is_current:
            raise RuntimeError("cannot join current thread")
        with self.cond:
            self.cond.wait(timeout)



    # this makes threading._after_fork work
    def __stopped(self, v):
        self.stopped = v
    __stopped = property(fset=__stopped)

    # required for threading._pickSomeNonDaemonThread
    def is_alive(self):
        return self.evt_started.is_set() and not self.stopped

    alive = property(is_alive)

    # fix for mext.log
    def getName(self):
        return self.name


_MainThread.is_main = True












from collections import deque
import heapq


class Empty(Exception):
    "Exception raised by Queue.get(block=0)."
    pass

class Full(Exception):
    "Exception raised by Queue.put(block=0)"
    pass

class Queue(object):
    def __init__(self, maxsize=0):
        self.maxsize = maxsize
        self._init(maxsize)
        self.mutex = Lock()
        # Notify not_empty whenever an item is added to the queue; a
        # thread waiting to get is notified then.
        self.not_empty = Condition(self.mutex)
        # Notify not_full whenever an item is removed from the queue;
        # a thread waiting to put is notified then.
        self.not_full = Condition(self.mutex)
        # Notify _cond_tasks whenever the number of unfinished tasks
        # drops to zero; thread waiting to join() is notified to resume
        self._cond_tasks = Condition(self.mutex)
        self.unfinished_tasks = 0

    def task_done(self):
        with self._cond_tasks:
            v = self.unfinished_tasks - 1
            if v <= 0:
                if v < 0:
                    raise ValueError('task_done() called too many times')
                self._cond_tasks.notify_all()
            self.unfinished_tasks = v

    def join(self):
        """Blocks until all items in the Queue have been gotten and processed.

        The count of unfinished tasks goes up whenever an item is added to the
        queue. The count goes down whenever a consumer thread calls task_done()
        to indicate the item was retrieved and all work on it is complete.

        When the count of unfinished tasks drops to zero, join() unblocks.
        """
        with self._cond_tasks:
            while self.unfinished_tasks:
                self._cond_tasks.wait()

    def __len__(self):
        return len(self.queue)

    @property
    def is_full(self):
        return (0 < self.maxsize == len(self.queue))

    def put(self, item, block=True, timeout=None):
        """Put an item into the queue.

        If optional args 'block' is true and 'timeout' is None (the default),
        block if necessary until a free slot is available. If 'timeout' is
        a positive number, it blocks at most 'timeout' seconds and raises
        the Full exception if no free slot was available within that time.
        Otherwise ('block' is false), put an item on the queue if a free slot
        is immediately available, else raise the Full exception ('timeout'
        is ignored in that case).
        """
        with self.not_full:
            if self.maxsize > 0:
                if not block:
                    if len(self.queue) == self.maxsize:
                        raise Full
                elif timeout is None:
                    while len(self.queue) == self.maxsize:
                        self.not_full.wait()
                elif timeout < 0:
                    raise ValueError("'timeout' must be a positive number")
                else:
                    endtime = time() + timeout
                    while len(self.queue) == self.maxsize:
                        remaining = endtime - time()
                        if remaining <= 0.0:
                            raise Full
                        self.not_full.wait(remaining)
            self._put(item)
            self.unfinished_tasks += 1
            self.not_empty.notify()


    def get(self, block=True, timeout=None):
        """Remove and return an item from the queue.

        If optional args 'block' is true and 'timeout' is None (the default),
        block if necessary until an item is available. If 'timeout' is
        a positive number, it blocks at most 'timeout' seconds and raises
        the Empty exception if no item was available within that time.
        Otherwise ('block' is false), return an item if one is immediately
        available, else raise the Empty exception ('timeout' is ignored
        in that case).
        """
        with self.not_empty:
            if not block:
                if not len(self.queue):
                    raise Empty
            elif timeout is None:
                while not len(self.queue):
                    self.not_empty.wait()
            elif timeout < 0:
                raise ValueError("'timeout' must be a positive number")
            else:
                endtime = time() + timeout
                while not len(self.queue):
                    remaining = endtime - time()
                    if remaining <= 0.0:
                        raise Empty
                    self.not_empty.wait(remaining)
            item = self._get()
            self.not_full.notify()
            return item

    def _init(self, maxsize):
        q = self.queue = deque()
        self._put = q.append
        self._get = q.popleft


class PriorityQueue(Queue):
    def _init(self, maxsize):
        self.queue = []

    def _put(self, item, heappush=heapq.heappush):
        heappush(self.queue, item)

    def _get(self, heappop=heapq.heappop):
        return heappop(self.queue)


class LifoQueue(Queue):
    def _init(self, maxsize):
        q = self.queue = []
        self._put = q.append
        self._get = q.pop






















## t = Thread(lambda: x)
## t.ignore_errors = True
## t.start()
##
##
## @Thread
## def loop():
##     while True:
##         pass
##
## #loop.e = sys.stderr
## loop.start_daemon()
##

## C:\Work\Apps\S3 Backup\planning.py:12:from Queue import PriorityQueue, Queue, Empty
## C:\Work\Modules\Mext\mext\app\wx.py:10:from Queue import Queue
## C:\Work\Modules\Mext\mext\app\wx.py:82:            q = Queue()



## class ThreadExt(object):
##     #
##     # thread killing
##     #
##
##     @property
##     def can_kill(self):
##         try:
##             from ctypes import pythonapi
##             return True
##         except ImportError:
##             return False
##
##     def kill(self):
##         self.async_raise(SystemExit)
##
##     def async_raise(self, exctype):
##         import ctypes
##         #thread.get_ident()
##         if not issubclass(exctype, BaseException):
##             raise TypeError("Only types can be raised (not instances)")
##         if not isinstance(self.id, int):
##             raise TypeError("tid must be an integer")
##         res = ctypes.pythonapi.PyThreadState_SetAsyncExc(self.id, ctypes.py_object(exctype))
##         if res == 0:
##             raise ValueError("invalid thread id")
##         elif res != 1:
##             # """if it returns a number greater than one, you're in trouble,
##             # and you should call it again with exc=NULL to revert the effect"""
##             ctypes.pythonapi.PyThreadState_SetAsyncExc(self.id, 0)
##             raise SystemError("PyThreadState_SetAsyncExc failed")
##

#@@ Thread.log = Logger('threads[$name/$id]', threads_log)














